/*
	TaltOS
	on the Philips Chip LPC2106

	General Public Licence
	By Peter Eordegh
*/

#include "lpc2106_memorymap.h"
#include "lpc2106_functions.h"
#include "TaltOS_defs.h"
#include "TaltOS.h"

typedef struct
{
	unsigned long* address;
	unsigned char torun;
	unsigned char priority;
	unsigned char ID;
	unsigned char state;
	unsigned long argument;
} TaltOS_task;

typedef struct
{
	unsigned long ID1;
	unsigned long ID2;
	unsigned long* address;
	unsigned long size;
} TaltOS_reg_entity;

typedef struct
{
	unsigned short input_buf_start;
	unsigned short input_buf_end;
	unsigned short input_buf_act_size;
	unsigned short output_buf_start;
	unsigned short output_buf_end;
	unsigned short output_buf_act_size;
	TaltOS_task actual_task;
} TaltOS_env;

void TaltOS_clear_tasklist()
{
	TaltOS_task* temp;
	unsigned int i;
	for(i = 0; i < TALTOS_TASK_LIST_SIZE; i++)
	{
		temp = ((TaltOS_task*)(TALTOS_TASK_LIST_PTR)) + i;
		temp->address  = 0;
		temp->torun    = 0;
		temp->priority = 0;
		temp->ID       = 0;
		temp->state    = 0;
		temp->argument = 0;
	}
}

//Clear the flash operation buffer
void TaltOS_clear_flash_op_buffer()
{
	unsigned int i;
	for(i = 0; i < (TALTOS_FLASH_OP_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR + i)) = (unsigned long)TALTOS_EMPTY;
	}
}


//Clear the input buffer
void TaltOS_clear_input_buffer()
{
	unsigned int i;
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	for(i = 0; i < (TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_INPUT_BUFFER_PTR + i)) = (unsigned long)TALTOS_EMPTY;
	}
	env->input_buf_start = 0;
	env->input_buf_end = 0;
	env->input_buf_act_size = 0;
}

//Clear the output buffer
void TaltOS_clear_output_buffer()
{
	unsigned int i;
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	for(i = 0; i < (TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_INPUT_BUFFER_PTR + i)) = (unsigned long)TALTOS_EMPTY;
	}
	env->output_buf_start = 0;
	env->output_buf_end = 0;
	env->output_buf_act_size = 0;
}

//Clear the environement variables of the TalOS
void TaltOS_clear_environement()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	env->input_buf_start	= 0;
	env->input_buf_end		= 0;
	env->input_buf_act_size = 0;
	env->output_buf_start	= 0;
	env->output_buf_end		= 0;
	env->output_buf_act_size = 0;

	env->actual_task.address  = 0;
	env->actual_task.torun    = 0;
	env->actual_task.priority = 0;
	env->actual_task.ID       = 0;
	env->actual_task.state    = 0;
	env->actual_task.argument = 0;
}

//The TaltOS initialization routine
void TaltOS_init()
{
	TaltOS_clear_flash_op_buffer();
	TaltOS_clear_input_buffer();
	TaltOS_clear_output_buffer();
	TaltOS_clear_environement();
	//LPC2106 specific code
	REG8(TIMER0_TCR) = 0x1;
	lpc2106_vectored_interrupts_setup();
	//End of LPC2106 specific code
	TaltOS_run_components_init();
	TaltOS_loop();
}

//The TaltOS main loop
void TaltOS_loop()
{
	unsigned long counter = 0;
	while(1)
	{
		if (counter < 100000) {counter++;}
		else
		{
			counter = 0;
		}
	}
}

//Run all the components' init function (needed after a reset)
void TaltOS_run_components_init()
{
	unsigned int i;
	char* temp;
	unsigned int j;
	TaltOS_reg_entity *comp_reg;
	void (*comp_init)();
	for(i = 0; i < TALTOS_COMPONENT_TABLE_SIZE/sizeof(TaltOS_reg_entity); i++)
	{
		comp_reg = (TaltOS_reg_entity*)TALTOS_COMPONENT_TABLE_PTR+i;
		if((comp_reg->ID1!=TALTOS_EMPTY) && (comp_reg->ID2!=TALTOS_EMPTY))
		{
			temp=(char*)comp_reg;
			TaltOS_char_output_write(' ');
			for(j = 0; j < 8; j++)
			{
				if(j==4){TaltOS_char_output_write(':');}
				TaltOS_char_output_write(*(temp+j)+'0');
			}
			if(
				!((comp_reg->ID1==TALTOS_ERASED) && (comp_reg->ID2==TALTOS_ERASED)) &&
			    !((comp_reg->ID1==TALTOS_ID1) && (comp_reg->ID2==TALTOS_ID2)) &&
			    !(TaltOS_is_garbage_reg((char*)comp_reg))
			    )
			{
				comp_init =(void(*)())(comp_reg->address+(*(comp_reg->address))/4);
				comp_init();
			}
		}
	}
}


//The TaltOS tic routine for the preemtive task change
void TaltOS_tic()
{
	//REG8(UART0_THR) = 'p';
}

//The input buffer's input function, can be called by the microcontroller
void TaltOS_char_input_write(unsigned char inchar)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	char* address = (char*)(TALTOS_INPUT_BUFFER_PTR + env->input_buf_end);
	(*address) = inchar;

	if (env->input_buf_end < TALTOS_INPUT_BUFFER_SIZE)
	{
		if (env->input_buf_start != env->input_buf_end + 1){env->input_buf_end++;}
	}
	else
	{
		if (env->input_buf_start != 0){env->input_buf_end = 0;}
	}

	(env->input_buf_act_size)++;

	if(TaltOS_component_arrived())
	{
		TaltOS_char_output_write('C');
	}

}


//The input buffer's output function, can be called by the tasks to recive data from out of the system
int TaltOS_char_input_read(unsigned char* ret)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	if (env->input_buf_start == env->input_buf_end){return 0;}
	(*ret) = (*((char*)(TALTOS_INPUT_BUFFER_PTR + env->input_buf_start)));

	if (env->input_buf_start < TALTOS_INPUT_BUFFER_SIZE) {env->input_buf_start++;}
	else {env->input_buf_start = 0;}
	(env->input_buf_act_size)--;
	return 1;
}

//The output buffer's input function, can be called by the tasks to send data out of system
void TaltOS_char_output_write(unsigned char inchar)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	char* address = (char*)(TALTOS_OUTPUT_BUFFER_PTR + env->output_buf_end);
	(*address) = inchar;

	if (env->output_buf_end < TALTOS_OUTPUT_BUFFER_SIZE)
	{
		if (env->output_buf_start != env->output_buf_end + 1){env->output_buf_end++;}
	}
	else
	{
		if (env->output_buf_start != 0){env->output_buf_end = 0;}
	}

	(env->output_buf_act_size)++;

	//LPC2106 specific code
	lpc2106_force_uart_irq();
	//LPC2106 specific code end
}

//The output buffer's output function, can be called by the microcontroller
int TaltOS_char_output_read(unsigned char* ret)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	if (env->output_buf_start == env->output_buf_end){return 0;}
	(*ret) = (*((char*)(TALTOS_OUTPUT_BUFFER_PTR + env->output_buf_start)));

	if (env->output_buf_start < TALTOS_OUTPUT_BUFFER_SIZE) {env->output_buf_start++;}
	else {env->output_buf_start = 0;}
	(env->output_buf_act_size)--;
	return 1;
}


//Get the component's address by it's ID (Absolute address)
unsigned long* TaltOS_get_component_address(unsigned long comp_id1, unsigned long comp_id2)
{
	unsigned int isfound = 0;
	TaltOS_reg_entity *comp_reg;
	unsigned int i;
	for(i = 0; (i < TALTOS_COMPONENT_TABLE_SIZE/sizeof(TaltOS_reg_entity)) && (!isfound); i++)
	{
		comp_reg = (TaltOS_reg_entity*)TALTOS_COMPONENT_TABLE_PTR+i;
		if ((comp_reg->ID1 == comp_id1) && (comp_reg->ID2 == comp_id2))
		{
			isfound = 1;
		}
	}
	if (isfound) {return comp_reg->address;}
	else {return 0;}
}

//Get the component's n. function's pointer (Absolute address)
unsigned long TaltOS_get_component_function(unsigned long* address, unsigned long index)
{
	return (unsigned long)(address+(*(address+index))/4);
}

//Check if a component has arrived into the input buffer
unsigned int TaltOS_component_arrived()
{
	unsigned int i;
	unsigned int ret=1;
	short tempend;
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	if(env->input_buf_act_size>5)
	{
		tempend=env->input_buf_end;
		for(i = 0; i < 5; i++)
		{
			if (tempend > 0) {tempend--;}
			else {tempend = TALTOS_INPUT_BUFFER_SIZE-1;}
			switch (i)
			{
				case 0 : if (*((char*)(TALTOS_INPUT_BUFFER_PTR+tempend))!=TALTOS_COMP_END_5){ret=0;}break;
				case 1 : if (*((char*)(TALTOS_INPUT_BUFFER_PTR+tempend))!=TALTOS_COMP_END_4){ret=0;}break;
				case 2 : if (*((char*)(TALTOS_INPUT_BUFFER_PTR+tempend))!=TALTOS_COMP_END_3){ret=0;}break;
				case 3 : if (*((char*)(TALTOS_INPUT_BUFFER_PTR+tempend))!=TALTOS_COMP_END_2){ret=0;}break;
				case 4 : if (*((char*)(TALTOS_INPUT_BUFFER_PTR+tempend))!=TALTOS_COMP_END_1){ret=0;}break;
			}
			if (ret==0) {break;}
		}
		return ret;
	}

	return 0;

}


//Checks if the component's ids are garbage (wrong value => error in erase operation)
unsigned int TaltOS_is_garbage_reg(char *test_array)
{
	if (test_array[0]==0x4E && test_array[4]==TALTOS_COMPID_END) {return 0;}
	else return 1;
}

/*void TaltOS_copy_component(unsigned long id1, unsigned long id2, unsigned long size, unsigned char* start_addr)
{

	unsigned long* empty_tester;
	bool address_found=false;
	unsigned long i,j;
	TaltOS_reg_entity *TaltOS_reg;
	//Get the size of the buffer to be written
	TaltOS.size = (buffer_pointer/LPC2106_MINDATA)+1;

	//Search for the first empty place of the TaltOS
	for(i=0; (i<LPC2106_FLASH_MEMORY_SIZE/LPC2106_MINDATA) && (!address_found); i++)
	{
		address_found=true;
		// Check if the full size is fit to this place
		for(j=0; (j<TaltOS.size)&&(address_found); j++)
		{
			empty_tester=(unsigned long*)((i+j)*LPC2106_MINDATA);
			if ((*empty_tester)!=TALTOS_EMPTY)
			{
				address_found=false;
			}
		}
	}

	if (address_found)
	{
		TaltOS.address = (unsigned long*)((i-1)*LPC2106_MINDATA);

		//Copy the buffer by LPC2106_MINDATA size blocks
		for(j=0; j<TaltOS.size; j++)
		{
			copy_to_flash_512(	(unsigned long)TaltOS.address+j*LPC2106_MINDATA,
								(unsigned long*)(TALTOS_INPUT_BUFFER_PTR+j*LPC2106_MINDATA),
								1);
		}

		clear_input_buffer();

		//Save the full component table, and register the TaltOS if possible
		address_found=false;
		for(i = 0; i < (TALTOS_FLASH_OP_BUFFER_SIZE/sizeof(TaltOS_reg_entity)); i++)
		{
			(*((TaltOS_reg_entity*)(TALTOS_FLASH_OP_BUFFER_PTR)+i)) = (*((TaltOS_reg_entity*)(TALTOS_COMPONENT_TABLE_PTR)+i));
			TaltOS_reg = (TaltOS_reg_entity*)(TALTOS_FLASH_OP_BUFFER_PTR)+i;
			if ((!address_found)&&(TaltOS_reg->ID1 == TALTOS_EMPTY)&&(TaltOS_reg->ID2 == TALTOS_EMPTY))
			{
				TaltOS_reg->ID1=TaltOS.ID1;
				TaltOS_reg->ID2=TaltOS.ID2;
				TaltOS_reg->address=TaltOS.address;
				TaltOS_reg->size=TaltOS.size;
				address_found=true;
			}
		}

		//Copy the full component table back to the flash
		copy_to_flash_512(TALTOS_COMPONENT_TABLE_PTR,(unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR,16);

		clear_flash_op_buffer();

		//Run the TaltOS
		what = 1;
	}
	else
	{
		REG8(UART0_THR) = 'F';		//F as Fail
		what = 0;
	}
}*/
